/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * Portions of this software are based upon public domain software
 * originally written at the National Center for Supercomputing Applications,
 * University of Illinois, Urbana-Champaign.
 */

#define CORE_PRIVATE
#include "apr.h"
#include "ap_config.h"
#include "httpd.h"
#include "ap_iol.h"
#include "apr_network_io.h"
#include "apr_file_io.h"
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#define FD_NONBLOCKING_SET	(1)

typedef struct {
    ap_iol iol;
    apr_socket_t *sock;
} iol_socket;

static apr_status_t socket_setopt(ap_iol *viol, ap_iol_option opt,
                               const void *value)
{
    iol_socket *iol = (iol_socket *)viol;

    switch (opt) {
    case AP_IOL_TIMEOUT:
        apr_setsocketopt(iol->sock, APR_SO_TIMEOUT, 
                        *(const int*)value * AP_USEC_PER_SEC);
        break;
    default:
	    return APR_EINVAL;
    }
    return APR_SUCCESS;
}

static apr_status_t socket_getopt(ap_iol *viol, ap_iol_option opt, void *value)
{
    iol_socket *iol = (iol_socket *)viol;

    switch (opt) {
    case AP_IOL_TIMEOUT:
        apr_getsocketopt(iol->sock, APR_SO_TIMEOUT, (apr_int32_t *) value);
        (*(apr_int32_t *)value) /= AP_USEC_PER_SEC;
        break;
    default:
        return APR_EINVAL;
    }
    return APR_SUCCESS;
}


static apr_status_t socket_write(ap_iol *viol, const char* buf, apr_size_t size,
     apr_ssize_t *nbytes)
{
    *nbytes = size;
    return apr_send(((iol_socket *)viol)->sock, buf, nbytes);
}

static apr_status_t socket_writev(ap_iol *viol, const struct iovec *vec, apr_size_t nvec,
    apr_ssize_t *nbytes)
{
#ifndef HAVE_NETINET_TCP_H
    /* At present BeOS doesn't support writev for sockets... */
    *nbytes = vec[0].iov_len;
    return apr_send(((iol_socket *)viol)->sock, vec[0].iov_base, nbytes);
#else
    return apr_sendv(((iol_socket *)viol)->sock, vec, nvec, nbytes);
#endif
}

static apr_status_t socket_read(ap_iol *viol, char* buf, apr_size_t size,
     apr_ssize_t *nbytes)
{
    *nbytes = size;
    return apr_recv(((iol_socket *)viol)->sock, buf, nbytes);
}

#if APR_HAS_SENDFILE
static apr_status_t socket_sendfile(ap_iol *viol, apr_file_t *file,
                                   apr_hdtr_t *hdtr, apr_off_t *offset, 
                                   apr_size_t *len,  apr_int32_t flags) 
{
    return apr_sendfile(((iol_socket *)viol)->sock, file, hdtr, offset, len, flags);
}
#endif

static apr_status_t socket_close(ap_iol *viol)
{
    iol_socket *iol = (iol_socket *)viol;
    int saved_errno;

    saved_errno = apr_close_socket(iol->sock);
    iol = NULL;
    if (saved_errno == 0) {
        return APR_SUCCESS;
    }
    return saved_errno;
}

static apr_status_t socket_shutdown(ap_iol *viol, int how)
{
    iol_socket *iol = (iol_socket *)viol;

    return apr_shutdown(iol->sock, how);
}
    
static const ap_iol_methods socket_methods = {
    socket_close,
    socket_write,
    socket_writev,
    socket_read,
    socket_setopt,
    socket_getopt,
#if APR_HAS_SENDFILE
    socket_sendfile,
#else
    NULL,
#endif
    socket_shutdown,
};

API_EXPORT(ap_iol *) ap_iol_attach_socket(apr_pool_t *p, apr_socket_t *sock)
{
    iol_socket *iol;

    iol = apr_palloc(p, sizeof(iol_socket));

    iol->iol.methods = &socket_methods;
    iol->sock = sock;
    return (ap_iol *)iol;
}

API_EXPORT(apr_socket_t *) ap_iol_get_socket(ap_iol *iol)
{
    iol_socket *realiol = (iol_socket *)iol;

    return realiol->sock;
}
