/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * Portions of this software are based upon public domain software
 * originally written at the National Center for Supercomputing Applications,
 * University of Illinois, Urbana-Champaign.
 */

#ifndef MOD_CACHE_H
#define MOD_CACHE_H 

#include "ap_cache.h"

#include "httpd.h"
#include "http_config.h"
#include "http_protocol.h"
#include "util_date.h"

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <sys/stat.h>

#define TESTING 0

extern module MODULE_VAR_EXPORT cache_module;

/* number of characters in the hash */
#define HASH_LEN (22*2)

/* maximum  'CacheDirLevels*CacheDirLength' value */
#define CACHEFILE_LEN 20    /* must be less than HASH_LEN/2 */

#ifdef CHARSET_EBCDIC
#define CRLF   "\r\n"
#else /*CHARSET_EBCDIC*/
#define CRLF   "\015\012"
#endif /*CHARSET_EBCDIC*/

#define CACHE_HEADER_SUFFIX ".header"
#define CACHE_DATA_SUFFIX ".data"

#define    MSEC_ONE_DAY        ((apr_time_t)(86400*AP_USEC_PER_SEC))    /* one day, in microseconds */
#define    MSEC_ONE_HR        ((apr_time_t)(3600*AP_USEC_PER_SEC))    /* one hour, in microseconds */

#define DEFAULT_CACHE_SPACE 5
#define DEFAULT_CACHE_MAXEXPIRE MSEC_ONE_DAY
#define DEFAULT_CACHE_EXPIRE    MSEC_ONE_HR
#define DEFAULT_CACHE_LMFACTOR (0.1)

/* static information about the local cache */
typedef struct {
    const char *root;           /* the location of the cache directory */
    off_t space;                /* Maximum cache size (in 1024 bytes) */
    apr_time_t maxexpire;        /* Maximum time to keep cached files in msecs */
    apr_time_t defaultexpire;    /* default time to keep cached file in msecs */
    double lmfactor;            /* factor for estimating expires date */
    apr_time_t gcinterval;       /* garbage collection interval, in msec */
    int dirlevels;              /* Number of levels of subdirectories */
    int dirlength;              /* Length of subdirectory names */
} cache_server_conf;

/* caching information about a request */
enum read_states { FC_READ_NONE=0x00, FC_READ_MY_DATA=0x01, FC_READ_HEADERS=0x02, FC_TEMP_FILE=0x04 };
typedef struct {
    ap_cache_el el;
    char *tempfile;          /* name of the temporary file, used for cache
                                element creation */

    /* states and gc states */
    char state;
    apr_time_t expire;        /* calculated expire date of cached entity */
    apr_time_t date;          /* the date the cached file was last touched */
    int version;             /* update count of the file */

    /* names of files */
    char *datafile;          /* where the data will go */
    char *hdrsfile;          /* where the hdrs will go */

    /* stuff the API uses */
    apr_table_t *hdrs;        /* the HTTP headers of the file */
    BUFF *fp;                /* the cache file descriptors */
} cache_req;

/* in mod_cache.c */
apr_status_t file_cache_close(ap_cache_handle_t *h);
apr_status_t file_cache_el_hdr_walk(ap_cache_el *el,
         int (*comp)(void *, const char *, const char *),
         void *rec, va_list args);
apr_status_t file_cache_el_hdr(ap_cache_el *el, const char *name,
         const char *val, ap_cache_query flag);
apr_status_t file_cache_el_data(ap_cache_el *el, BUFF **b);
apr_status_t file_cache_reset(ap_cache_el *el, ap_cache_part f);

/* in file_garbage.c */
apr_status_t file_cache_gc(ap_cache_handle_t *r);

/* in file_cache.c */
apr_status_t file_cache_element(ap_cache_handle_t *h, const char *name, ap_cache_el **el_in, ap_cache_query flag);
apr_status_t file_cache_el_final(ap_cache_el *el);

/* in cache_util.c */
char *header_file(ap_cache_handle_t *h, const char *name); /* generate hdrsfile */
int file_cache_write_hdrs(cache_req *el); /* write headers to file */
int file_cache_read_hdrs(cache_req *el); /* read headers from file */
int file_cache_get_hdrs(cache_req *el); /* prepare el's hdrs */
char *data_file(ap_cache_handle_t *h, const char *name); /* generate datafile */
int file_cache_get_data(cache_req *el); /* prepare el's fp */
int mkdir_structure(char *file, const char *root);
apr_time_t ap_cache_hex2msec(const char *x);
void ap_cache_msec2hex(apr_time_t j, char *y);

#endif
